#! /usr/bin/env python3

# This is a modified version of Pierre Christian and Chi-kwan Chan's
# FANTASY.py <https://github.com/pierrechristian/FANTASY>
# It was modified in 2021 by David Wright to work with the
# Kerr-Newman metric, resolve some ambiguities regarding
# the overloading of trigonometric functions, and reduce imported
# dependencies.
#
# Copyright (C) 2020 Pierre Christian and Chi-kwan Chan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''
################### USER GUIDE ###################
FANTASY is a geodesic integration code for arbitrary metrics with automatic differentiation. Please refer to Christian and Chan, 2021 for details.

################### Inputing the Metric ###################
Components of the metric are stored in the functions g00, g01, g02, etc that can be found under the heading "Metric Components". Each of these take as input a list called Param, which contains the fixed parameters of the metric (e.g., 'M' and 'a' for the Kerr metric in Boyer-Lindquist coordinates) and a list called Coord, which contains the coordinates (e.g., 'r' and 't' for the Kerr metric in Boyer-Lindquist coordinates). In order to set up a metric,
Step 1) Write down the fixed parameters in a list
Step 2) Write down the coordinates in a list
Step 3) Type the metric into the functions under "Metric Components".

Example: Kerr in Boyer-Lindquist coordinates
Step 1) The fixed parameters are listed as [M,a]
Step 2) The coordinates are listed as [t,r,theta,phi]
Step 3) Type in the metric components, for example, the g11 function becomes:

def g11(Param,Coord):
    return (Param[1]**2.-2.*Param[0]*Coord[1]+Coord[1]**2.)/(Coord[1]**2.+Param[1]**2.*cos(Coord[2])**2.)

Extra step) To make your code more readable, you can redefine variables in place of Param[i] or Coord[i], for example, the g11 function can be rewritten as:
def g11(Param,Coord):
    M = Param[0]
    a = Param[1]
    r = Coord[1]
    theta = Coord[2]
    return (a**2.-2.*M*r+r**2.)/(r**2.+a**2.*cos(theta)**2.)

################### A Guide on Choosing omega ###################
The parameter omega determines how much the two phase spaces interact with each other. The smaller omega is, the smaller the integration error, but if omega is too small, the equation of motion will become non-integrable. Thus, it is important to find an omega that is appropriate for the problem at hand. The easiest way to choose an omega is through trial and error:

Step 1) Start with omega=1; if you are working in geometric/code units in which all relevant factors are ~unity, this is usually already a good choice of omega
Step 2) If the trajectory varies wildly with time (this indicates highly chaotic, non-integrable behavior), increase omega and redo integration
Step 3) Repeat Step 2) until trajectory converges

################### Running the Code ###################
To run the code, run the function geodesic_integrator(N,delta,omega,q0,p0,Param,order). N is the number of steps, delta is the timestep, omega is the interaction parameter between the two phase spaces, q0 is a list containing the initial position, p0 is a list containing the initial momentum, Param is a list containing the fixed parameters of the metric (e.g., [M,a] for Kerr metric in Boyer-Lindquist coordinates), and order is the integration order. You can choose either order=2 for a 2nd order scheme or order=4 for a 4th order scheme.

################### Reading the Output ###################
The output is a numpy array indexed by timestep. For each timestep, the output contains four lists:

output[timestep][0] = a list containing the position of the particle at said timestep in the 1st phase space
output[timestep][1] = a list containing the momentum of the particle at said timestep in the 1st phase space
output[timestep][2] = a list containing the position of the particle at said timestep in the 2nd phase space
output[timestep][3] = a list containing the momentum of the particle at said timestep in the 2nd phase space

As long as the equation of motion is integrable (see section "A Guide on Choosing omega"), the trajectories in the two phase spaces will quickly converge, and you can choose either one as the result of your calculation.

################### Automatic Jacobian ###################

Input coordinate transformations for the 0th, 1st, 2nd, 3rd coordinate in functions CoordTrans0, CoordTrans1, CoordTrans2, CoordTrans3. As an example, coordinate transformation from Spherical Schwarzschild to Cartesian Schwarzschild has been provided.

'''

################### Code Preamble ###################

import numpy as np
from numpy import sin, cos
from IPython.display import clear_output, display

class dual:
  def __init__(self, first, second):
    self.f = first
    self.s = second

  def __mul__(self,other):
    if isinstance(other,dual):
      return dual(self.f*other.f, self.s*other.f+self.f*other.s)
    else:
      return dual(self.f*other, self.s*other)

  def __rmul__(self,other):
    if isinstance(other,dual):
      return dual(self.f*other.f, self.s*other.f+self.f*other.s)
    else:
      return dual(self.f*other, self.s*other)

  def __add__(self,other):
    if isinstance(other,dual):
      return dual(self.f+other.f, self.s+other.s)
    else:
      return dual(self.f+other,self.s)

  def __radd__(self,other):
    if isinstance(other,dual):
      return dual(self.f+other.f, self.s+other.s)
    else:
      return dual(self.f+other,self.s)

  def __sub__(self,other):
    if isinstance(other,dual):
      return dual(self.f-other.f, self.s-other.s)
    else:
      return dual(self.f-other,self.s)

  def __rsub__(self, other):
    return dual(other, 0) - self

  def __truediv__(self,other):
    ''' when the first component of the divisor is not 0 '''
    if isinstance(other,dual):
      return dual(self.f/other.f, (self.s*other.f-self.f*other.s)/(other.f**2.))
    else:
      return dual(self.f/other, self.s/other)

  def __rtruediv__(self, other):
    return dual(other, 0).__truediv__(self)

  def __neg__(self):
      return dual(-self.f, -self.s)

  def __pow__(self, power):
    return dual(self.f**power,self.s * power * self.f**(power - 1))

  def sin(self):
    return dual(np.sin(self.f),self.s*np.cos(self.f))

  def cos(self):
    return dual(np.cos(self.f),-self.s*np.sin(self.f))

  def tan(self):
    return self.sin()/self.cos()

  def log(self):
    return dual(np.log(self.f),self.s/self.f)

  def exp(self):
    return dual(np.exp(self.f),self.s*np.exp(self.f))

def dif(func,x):
    funcdual = func(dual(x,1.))
    if isinstance(funcdual,dual):
        return func(dual(x,1.)).s
    else:
        ''' this is for when the function is a constant, e.g. gtt:=0 '''
        return 0

################### Metric Components ###################

# Diagonal components of the metric
def g00(Param,Coord):
    M = Param[0]
    a = Param[1] / Param[0]
    Q = Param[2]
    r = Coord[1]
    theta = Coord[2]
    Delta = r**2.-2.*M*r+a**2. + Q**2
    rhosq = r**2.+a**2.*cos(theta)**2.
    return -(r**2.+a**2.+2.*M*r*a**2.*sin(theta)**2./rhosq)/Delta

def g11(Param,Coord):
    M = Param[0]
    a = Param[1] / Param[0]
    Q = Param[2]
    r = Coord[1]
    theta = Coord[2]
    Delta = r**2.-2.*M*r+a**2. + Q**2
    rhosq = r**2.+a**2.*cos(theta)**2.
    theta = Coord[2]
    return (Delta)/(rhosq)

def g22(Param,Coord):
    M = Param[0]
    a = Param[1] / Param[0]
    Q = Param[2]
    r = Coord[1]
    theta = Coord[2]
    theta = Coord[2]
    Delta = r**2.-2.*M*r+a**2. + Q**2
    rhosq = r**2.+a**2.*cos(theta)**2.
    return 1./rhosq

def g33(Param,Coord):
    M = Param[0]
    a = Param[1] / Param[0]
    Q = Param[2]
    r = Coord[1]
    theta = Coord[2]
    theta = Coord[2]
    Delta = r**2.-2.*M*r+a**2. + Q**2
    rhosq = r**2.+a**2.*cos(theta)**2.
    return (1./(Delta*sin(theta)**2.))*(1.-2.*M*r/rhosq)

# Off-diagonal components of the metric
def g01(Param,Coord):
    return 0

def g02(Param,Coord):
    return 0

def g03(Param,Coord):
    M = Param[0]
    a = Param[1]
    Q = Param[2]
    r = Coord[1]
    theta = Coord[2]
    Delta = r**2.-2.*M*r+a**2. + Q**2
    rhosq = r**2.+a**2.*cos(theta)**2.
    return -(2.*M*r*a)/(rhosq*Delta)

def g12(Param,Coord):
    return 0

def g13(Param,Coord):
    return 0

def g23(Param,Coord):
    return 0

################### Metric Derivatives ###################

def dm(Param,Coord,metric,wrt):
    ''' wrt = 0,1,2,3 '''
    point_d = Coord[wrt]

    point_0 = dual(Coord[0],0)
    point_1 = dual(Coord[1],0)
    point_2 = dual(Coord[2],0)
    point_3 = dual(Coord[3],0)

    if metric == 'g00':
        if wrt == 0:
            return dif(lambda p:g00(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g00(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g00(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g00(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g11':
        if wrt == 0:
            return dif(lambda p:g11(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g11(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g11(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g11(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g22':
        if wrt == 0:
            return dif(lambda p:g22(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g22(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g22(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g22(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g33':
        if wrt == 0:
            return dif(lambda p:g33(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g33(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g33(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g33(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g44':
        if wrt == 0:
            return dif(lambda p:g44(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g44(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g44(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g44(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g01':
        if wrt == 0:
            return dif(lambda p:g01(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g01(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g01(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g01(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g02':
        if wrt == 0:
            return dif(lambda p:g02(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g02(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g02(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g02(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g03':
        if wrt == 0:
            return dif(lambda p:g03(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g03(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g03(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g03(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g12':
        if wrt == 0:
            return dif(lambda p:g12(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g12(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g12(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g12(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g13':
        if wrt == 0:
            return dif(lambda p:g13(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g13(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g13(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g13(Param,[point_0,point_1,point_2,p]),point_d)
    elif metric == 'g23':
        if wrt == 0:
            return dif(lambda p:g23(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:g23(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:g23(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:g23(Param,[point_0,point_1,point_2,p]),point_d)

################### Automatic Coordinate Transformation ###################

def CoordTrans0(Param, Coord):

    M = Param[0]
    a = Param[1]
    t = Coord[0]
    
    return t
        
def CoordTrans1(Param, Coord):

    M = Param[0]
    a = Param[1]
    r = Coord[1]
    theta = Coord[2]
    phi = Coord[3]
    
    x = r*sin(theta)*cos(phi)

    return x

def CoordTrans2(Param, Coord):

    M = Param[0]
    a = Param[1]
    r = Coord[1]
    theta = Coord[2]
    phi = Coord[3]
    
    y = r*sin(theta)*sin(phi)

    return y

def CoordTrans3(Param, Coord):

    M = Param[0]
    a = Param[1]
    r = Coord[1]
    theta = Coord[2]
    
    z = r*cos(theta)

    return z

def AutoJacob(Param,Coord,i,wrt):
    
    point_d = Coord[wrt]

    point_0 = dual(Coord[0],0)
    point_1 = dual(Coord[1],0)
    point_2 = dual(Coord[2],0)
    point_3 = dual(Coord[3],0)

    if i == 0:
        if wrt == 0:
            return dif(lambda p:CoordTrans0(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:CoordTrans0(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:CoordTrans0(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:CoordTrans0(Param,[point_0,point_1,point_2,p]),point_d)

    if i == 1:
        if wrt == 0:
            return dif(lambda p:CoordTrans1(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:CoordTrans1(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:CoordTrans1(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:CoordTrans1(Param,[point_0,point_1,point_2,p]),point_d)

    if i == 2:
        if wrt == 0:
            return dif(lambda p:CoordTrans2(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:CoordTrans2(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:CoordTrans2(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:CoordTrans2(Param,[point_0,point_1,point_2,p]),point_d)

    if i == 3:
        if wrt == 0:
            return dif(lambda p:CoordTrans3(Param,[p,point_1,point_2,point_3]),point_d)
        elif wrt == 1:
            return dif(lambda p:CoordTrans3(Param,[point_0,p,point_2,point_3]),point_d)
        elif wrt == 2:
            return dif(lambda p:CoordTrans3(Param,[point_0,point_1,p,point_3]),point_d)
        elif wrt == 3:
            return dif(lambda p:CoordTrans3(Param,[point_0,point_1,point_2,p]),point_d)
    
        
################### Integrator ###################

def Hamil_inside(q,p,Param,wrt):
    return p[0]*p[0]*dm(Param,q,'g00',wrt) +  p[1]*p[1]*dm(Param,q,'g11',wrt) +  p[2]*p[2]*dm(Param,q,'g22',wrt) +  p[3]*p[3]*dm(Param,q,'g33',wrt) +  2*p[0]*p[1]*dm(Param,q,'g01',wrt) +  2*p[0]*p[2]*dm(Param,q,'g02',wrt) + 2*p[0]*p[3]*dm(Param,q,'g03',wrt) +  2*p[1]*p[2]*dm(Param,q,'g12',wrt) +  2*p[1]*p[3]*dm(Param,q,'g13',wrt) + 2*p[2]*p[3]*dm(Param,q,'g23',wrt)

def phiHA(delta,omega,q1,p1,q2,p2,Param):
    ''' q1=(t1,r1,theta1,phi1), p1=(pt1,pr1,ptheta1,pphi1), q2=(t2,r2,theta2,phi2), p2=(pt2,pr2,ptheta2,pphi2) '''
    dq1H_p1_0 = 0.5*(Hamil_inside(q1,p2,Param,0))
    dq1H_p1_1 = 0.5*(Hamil_inside(q1,p2,Param,1))
    dq1H_p1_2 =  0.5*(Hamil_inside(q1,p2,Param,2))
    dq1H_p1_3 =  0.5*(Hamil_inside(q1,p2,Param,3))

    p1_update_array = np.array([dq1H_p1_0,dq1H_p1_1,dq1H_p1_2,dq1H_p1_3])
    p1_updated = p1 - delta*p1_update_array

    dp2H_q2_0 = g00(Param,q1)*p2[0] + g01(Param,q1)*p2[1] + g02(Param,q1)*p2[2] + g03(Param,q1)*p2[3]
    dp2H_q2_1 = g01(Param,q1)*p2[0] + g11(Param,q1)*p2[1] + g12(Param,q1)*p2[2] + g13(Param,q1)*p2[3]
    dp2H_q2_2 = g02(Param,q1)*p2[0] + g12(Param,q1)*p2[1] + g22(Param,q1)*p2[2] + g23(Param,q1)*p2[3]
    dp2H_q2_3 = g03(Param,q1)*p2[0] + g13(Param,q1)*p2[1] + g23(Param,q1)*p2[2] + g33(Param,q1)*p2[3]

    q2_update_array = np.array([dp2H_q2_0,dp2H_q2_1,dp2H_q2_2,dp2H_q2_3])
    q2_updated = q2 + delta*q2_update_array

    return (q2_updated, p1_updated)

def phiHB(delta,omega,q1,p1,q2,p2,Param):
    ''' q1=(t1,r1,theta1,phi1), p1=(pt1,pr1,ptheta1,pphi1), q2=(t2,r2,theta2,phi2), p2=(pt2,pr2,ptheta2,pphi2) '''
    dq2H_p2_0 = 0.5*(Hamil_inside(q2,p1,Param,0))
    dq2H_p2_1 = 0.5*(Hamil_inside(q2,p1,Param,1))
    dq2H_p2_2 =  0.5*(Hamil_inside(q2,p1,Param,2))
    dq2H_p2_3 =  0.5*(Hamil_inside(q2,p1,Param,3))

    p2_update_array = np.array([dq2H_p2_0,dq2H_p2_1,dq2H_p2_2,dq2H_p2_3])
    p2_updated = p2 - delta*p2_update_array

    dp1H_q1_0 = g00(Param,q2)*p1[0] + g01(Param,q2)*p1[1] + g02(Param,q2)*p1[2] + g03(Param,q2)*p1[3]
    dp1H_q1_1 = g01(Param,q2)*p1[0] + g11(Param,q2)*p1[1] + g12(Param,q2)*p1[2] + g13(Param,q2)*p1[3]
    dp1H_q1_2 = g02(Param,q2)*p1[0] + g12(Param,q2)*p1[1] + g22(Param,q2)*p1[2] + g23(Param,q2)*p1[3]
    dp1H_q1_3 = g03(Param,q2)*p1[0] + g13(Param,q2)*p1[1] + g23(Param,q2)*p1[2] + g33(Param,q2)*p1[3]

    q1_update_array = np.array([dp1H_q1_0,dp1H_q1_1,dp1H_q1_2,dp1H_q1_3])
    q1_updated = q1 + delta*q1_update_array

    return (q1_updated, p2_updated)

def phiHC(delta,omega,q1,p1,q2,p2,Param):
    q1 = np.array(q1)
    q2 = np.array(q2)
    p1 = np.array(p1)
    p2 = np.array(p2)

    q1_updated = 0.5*( q1+q2 + (q1-q2)*np.cos(2.*omega*delta) + (p1-p2)*np.sin(2.*omega*delta) )
    p1_updated = 0.5*( p1+p2 + (p1-p2)*np.cos(2.*omega*delta) - (q1-q2)*np.sin(2.*omega*delta) )

    q2_updated = 0.5*( q1+q2 - (q1-q2)*np.cos(2.*omega*delta) - (p1-p2)*np.sin(2.*omega*delta) )
    p2_updated = 0.5*( p1+p2 - (p1-p2)*np.cos(2.*omega*delta) + (q1-q2)*np.sin(2.*omega*delta) )

    return (q1_updated, p1_updated, q2_updated, p2_updated)

def updator(delta,omega,q1,p1,q2,p2,Param):
    first_HA_step = np.array([q1, phiHA(0.5*delta,omega,q1,p1,q2,p2,Param)[1], phiHA(0.5*delta,omega,q1,p1,q2,p2,Param)[0], p2])
    first_HB_step = np.array([phiHB(0.5*delta,omega,first_HA_step[0],first_HA_step[1],first_HA_step[2],first_HA_step[3],Param)[0], first_HA_step[1], first_HA_step[2], phiHB(0.5*delta,omega,first_HA_step[0],first_HA_step[1],first_HA_step[2],first_HA_step[3],Param)[1]])
    HC_step = phiHC(delta,omega,first_HB_step[0],first_HB_step[1],first_HB_step[2],first_HB_step[3],Param)
    second_HB_step = np.array([phiHB(0.5*delta,omega,HC_step[0],HC_step[1],HC_step[2],HC_step[3],Param)[0], HC_step[1], HC_step[2], phiHB(0.5*delta,omega,HC_step[0],HC_step[1],HC_step[2],HC_step[3],Param)[1]])
    second_HA_step = np.array([second_HB_step[0], phiHA(0.5*delta,omega,second_HB_step[0],second_HB_step[1],second_HB_step[2],second_HB_step[3],Param)[1], phiHA(0.5*delta,omega,second_HB_step[0],second_HB_step[1],second_HB_step[2],second_HB_step[3],Param)[0], second_HB_step[3]])

    return second_HA_step

def updator_4(delta,omega,q1,p1,q2,p2,Param):
    z14 = 1.3512071919596578
    z04 = -1.7024143839193155
    step1 = updator(delta*z14,omega,q1,p1,q2,p2,Param)
    step2 = updator(delta*z04,omega,step1[0],step1[1],step1[2],step1[3],Param)
    step3 = updator(delta*z14,omega,step2[0],step2[1],step2[2],step2[3],Param)

    return step3

def geodesic_integrator(N,delta,omega,q0,p0,Param,order=2):
    q1=q0
    q2=q0
    p1=p0
    p2=p0

    result_list = [[q1,p1,q2,p2]]
    result = (q1,p1,q2,p2)

    for count, timestep in enumerate(range(N)):
        if order == 2:
            updated_array = updator(delta,omega,result[0],result[1],result[2],result[3],Param)
        elif order == 4:
            updated_array = updator_4(delta,omega,result[0],result[1],result[2],result[3],Param)

        result = updated_array
        result_list += [result]
        
        if not count%1000:
            clear_output(wait=True)
            display('On iteration number '+str(count) + ' with delta ' + str(delta))
        


    return result_list



#! /usr/bin/env python3

# Plot animated geodesics in 3D using hardware acceleration, possibly multiple at once.
# Copyright (C) 2021 David Wright
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
""

import argparse
import sys
from pathlib import Path
from typing import Union

import h5py
import numpy as np
import pyqtgraph.opengl as gl
from pyqtgraph.Qt import QtCore, QtGui


class Visualizer(object):
    """Visulization object

    Examples
    --------
    >>> v = Visualizer(['results1.h5', 'results2.h5'])
    >>> v.animation()

    """

    def __init__(self, results_files: Union[list, Path], schwarz_photon_sphere = True, mass=1):

        # Make sure we are working with a list of files, important for multiple geodesics at once
        if not isinstance(results_files, list):
            results_files = [results_files]
        self.results_files = results_files
        self.do_photon_sphere = schwarz_photon_sphere
        self.mass = mass

        # Some initializations for later
        self.traces = dict()
        self.app = QtGui.QApplication(sys.argv)
        self.w = gl.GLViewWidget()
        self.w.opts['distance'] = 80
        self.w.setWindowTitle('Black Hole Geodesics')
        self.w.setGeometry(0, 110, 1920, 1080)
        self.w.show()

        # Create the coordinate axes
        orientation = gl.GLAxisItem()
        orientation.setSize(x=10,y=10,z=10)
        self.w.addItem(orientation)


        # Do we want to see a schwarzschild photon sphere for a reference?
        if self.do_photon_sphere:
            sphere_mesh = gl.MeshData.sphere(rows=10, cols=10, radius=3*self.mass)
            self.photon_sphere = gl.GLMeshItem(meshdata=sphere_mesh, smooth=True,color=(1,1,1,0.4),
                                           shader='shaded',glOptions="translucent")
            self.photon_sphere.translate(0, 0, 0)
            self.w.addItem(self.photon_sphere)

        # Go through each file given and get the cartesian coords
        for result_file in results_files:
            with h5py.File(result_file, 'r') as hf:

                # I use the try except structure so that this works even for a single geodesic
                try:
                    self.t = np.column_stack((self.t, hf['time'][:]))
                except AttributeError:
                    self.t = hf['time'][:]
                except ValueError:
                    print("These simulations do not have matching dimensions!\nThe product of `num_steps` and `time_step` must be the same!")
                    sys.exit(-1)
                try:
                    self.x = np.column_stack((self.x, hf['x'][:]))
                except AttributeError:
                    self.x = hf['x'][:]
                try:
                    self.y = np.column_stack((self.y, hf['y'][:]))
                except AttributeError:
                    self.y = hf['y'][:]
                try:
                    self.z = np.column_stack((self.z, hf['z'][:]))
                except AttributeError:
                    self.z = hf['z'][:]

        # Get the points into the form pyqtgraph expects and plot them
        pts = np.vstack([self.x[0], self.y[0], self.z[0]]).transpose()
        self.traces = gl.GLScatterPlotItem(pos=pts)
        self.w.addItem(self.traces)


    def start(self):
        if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
            QtGui.QApplication.instance().exec_()

    def set_plotdata(self, name, points, color, width):
        self.traces[name].setData(pos=points) #, color=color, width=width)

    def update(self):
        """Update the plot

        Roll the coordinate arrays so that the next item in the array becomes
        the first. Set this item as the current displayed point.
        """

        self.x = np.roll(self.x,-10,axis=0)
        self.y = np.roll(self.y,-10,axis=0)
        self.z = np.roll(self.z,-10,axis=0)
        pts = np.vstack([self.x[0], self.y[0], self.z[0]]).transpose()
        self.traces.setData(pos=pts)

    def animation(self):
        timer = QtCore.QTimer()
        timer.timeout.connect(self.update)
        timer.start(20)
        self.start()


def cli():
    """Command line interface for animated 3D plotting module. """

    parser = argparse.ArgumentParser()
    # Accept one or more arguments
    parser.add_argument('datapath', nargs='+',
                        type=Path,
                        help="The path(s) to the data file(s).")
    args = parser.parse_args()

    v = Visualizer(args.datapath)
    v.animation()

# Start Qt event loop unless running in interactive mode.
if __name__ == '__main__':
    cli()



#!/usr/bin/env python3

# Evaluate geodesics for different spacetime backgrounds, possibly multiple in parallel.
# Copyright (C) 2021 David Wright
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import pickle as pl
from datetime import datetime as dt
from multiprocessing import Pool, cpu_count
from pathlib import Path
from shutil import copy2

import h5py
import matplotlib.pyplot as plt
import numpy as np
from yaml import safe_load

from event_horyzen.fantasy import geodesic_integrator


def run(conf_path=Path(__file__).parent / 'config.yml'):

    # Make the path to config file a list if it is not so that it works with the multiprocessing
    # implementation
    if not isinstance(conf_path,list):
        conf_path = [conf_path]



    # Empty lists to store values during multiprocessing
    args_list = []
    filename_list = []
    out_dir_list = []
    a_list = []

    # Loop over each provided configuration file
    for datapath in conf_path:
        with open(datapath, "r") as f:
            conf = safe_load(f)

            metric = conf['background_choice'].lower()
            bh_params = conf[metric]
            bh_params_list = list(bh_params.values())

        particle_params = conf['test_particle']

        # Initial 4-position and 4-momentum
        q0 = particle_params['q0']
        p0 = particle_params['p0']

        # Simulation parameters
        num_steps = conf['num_steps']
        time_step = conf['time_step']
        integration_order = conf['integration_order']
        omega = conf['omega']

        # The arguments are stored as a tuple so they can be mapped using Pool.starmap()
        # An example argument tuple with added information is shown below
        #
        # (N=1000, delta=0.5, omega=1, q0=q0, p0=p0, Param = bh_params_list, order=2 )]
        #
        # When you create your own argument tuple, you need to drop the keywords and just use
        # positional arguments
        args = (num_steps, time_step, omega, q0, p0, bh_params_list, integration_order )
        args_list.append(args)

        # Append information about each simulation
        filename_list.append(datapath)
        out_dir_list.append(conf['output_dir'])

        a_list.append(bh_params_list[1] / bh_params_list[0])

    # Start a multiprocessing pool and use as many cores as available to run the requested simulations
    with Pool(cpu_count()) as pool:
        results = pool.starmap(geodesic_integrator, args_list)

    results = np.array(results)

    # Colons are not present in timestamp because they are not allowed in Windows filenames!
    dt_string = dt.now().strftime("%Y-%m-%dT%H%M%S")

    # Iterate over each simulation and act
    for i, sim in enumerate(filename_list):

        # Get a unique output directory
        out_dir = Path(out_dir_list[i])
        sim_dir = Path(dt_string + "_" + sim.stem)
        p = out_dir / sim_dir
        p.mkdir(parents=True, exist_ok=True)

        # Copy the config file to the output directory so this simulation is reproducible
        copy2(sim,p)

        # We'll make a basic plot
        fig = plt.figure()
        ax = plt.axes(projection='3d')

        # Read in Boyer-Lindquist coords
        t = results[i,:,0,0]
        r = results[i,:,0,1]
        theta = results[i,:,0,2]
        phi = results[i,:,0,3]

        # We need the Kerr parameter in order to convert to cartesian from Boyer-Lindquist coords
        a = a_list[i]

        # Get cartesian coords for plotting
        x = np.sqrt(r**2 + a**2) * np.sin(theta) * np.cos(phi)
        y = np.sqrt(r**2 + a**2) * np.sin(phi) * np.sin(theta)
        z = r * np.cos(theta)

        ax.scatter3D(x, y, z,c=t, s =1)
        plt.savefig(p / 'basic-plot.png')

        # Pickle the plot so it can be interactively viewed later
        with open(p/'basic-plot.pickle', 'wb') as f:
            pl.dump(fig, f)

        # h5 files show significant file size reductions, and this is very
        # evident for N>10,000 simulations. If you don't care, see the
        # numpy.savetxt block below.
        with h5py.File(p/'results.h5', 'w') as hf:
            hf.create_dataset("time", data=t)
            hf.create_dataset("radius", data=r)
            hf.create_dataset("theta", data=theta)
            hf.create_dataset("phi", data=phi)
            hf.create_dataset("x", data=x)
            hf.create_dataset("y", data=y)
            hf.create_dataset("z", data=z)

        print("Output saved in", p)

        # If you prefer not to work with h5 files, uncomment this block and comment out the h5py block above
        """
        np.savetxt(p/'t.txt',t, header='Time values')
        np.savetxt(p/'r.txt',r, header='Radius values')
        np.savetxt(p/'theta.txt',theta, header='Theta values')
        np.savetxt(p/'phi.txt',phi, header='Phi values')
        np.savetxt(p/'x.txt',x, header='Cartesian X values')
        np.savetxt(p/'y.txt',y, header='Cartesian Y values')
        np.savetxt(p/'z.txt',z, header='Cartesian Z values')
        np.save(p/'results.npy', results)
        """
def copy_default_config(dest=Path.cwd()):
    copy2(Path(__file__).parent / 'config.yml', dest)
    print("Copied default configuration to {}".format(dest))



def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument('datapath', nargs='*',
                        default=Path(__file__).parent / 'config.yml',
                        type=Path,
                        help="The path(s) to the configuration file(s). Defaults to the included `config.yml` if not provided.")
    args = parser.parse_args()

    run(args.datapath)

if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['event_horyzen']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'h5py>=3.6.0,<4.0.0',
 'ipython>=7.30.1,<8.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.21.5,<2.0.0',
 'scipy>=1.7.3,<2.0.0']

extras_require = \
{'pyqt': ['pyqtgraph>=0.12.3,<0.13.0',
          'PyOpenGL>=3.1.5,<4.0.0',
          'PyQt5>=5.15.6,<6.0.0']}

entry_points = \
{'console_scripts': ['event-horyzen = event_horyzen.event_horyzen:cli',
                     'event-horyzen-plot = event_horyzen.animated_plot:cli']}

setup_kwargs = {
    'name': 'event-horyzen',
    'version': '0.1.2',
    'description': 'Simulates geodesic motion around black holes.',
    'long_description': '\n# Table of Contents\n\n1.  [Installation](#org9526257)\n    1.  [pip](#org8d3b001)\n    2.  [Manually](#org49b83bc)\n2.  [Usage](#org8d40f3c)\n    1.  [Example](#orgb8fd713)\n\n**Event hoRyzen** is a Python library designed to simulate and visualize geodesic motion around Schwarzschild, Reisner-Nordstrom, Kerr, and Kerr-Newman black holes.\nIt uses a slightly modified version of Pierre Christian and Chi-kwan Chan&rsquo;s FANTASY geodesic integration code (see <https://github.com/pierrechristian/FANTASY> + Pierre Christian and Chi-kwan Chan 2021 *ApJ* **909** 67).\n\n\n<a id="org9526257"></a>\n\n# Installation\n\n\n<a id="org8d3b001"></a>\n\n## pip\n\n    pip install event-horyzen\n\n**or**\n\n    pip install event-horzyen[pyqt]\n\n(If you use `zsh` as your shell you will need to quote the package name for optional dependencies, i.e., &ldquo;event-horyzen[pyqt]&rdquo;)\n\nDepending on whether or not you&rsquo;d like to use the `pyqtgraph` and `opengl` plotting modules (They are not small dependencies. The option to plot with matplotlib is included in the base package).\n\n\n<a id="org49b83bc"></a>\n\n## Manually\n\n    git clone https://github.com/UCF-SPS-Research-21/event-horyzen\n\nIf you use Poetry for package and venv management, you can use\n\n    poetry install\n\n**or**\n\n    poetry install -E pyqt\n\nIf you don&rsquo;t, you can `pip install -r requirements.txt` or `conda install --file requirements.txt`.\nThere are multiple versions of requirements.txt provided, it should be evident what each is for.\n\n\n<a id="org8d40f3c"></a>\n\n# Usage\n\nThe code is configured with a YAML configuration file.\nPlease see the example at <event_horyzen/config.yml>\n\n\n<a id="orgb8fd713"></a>\n\n## Example\n\n**I use Unix paths in the examples. Windows paths will work too &#x2014; just make sure you escape backslashes or make use of `pathlib`&rsquo;s functionality.**\n\nIf you&rsquo;d like to use the default configuration, you can just leave the argument to `event_horyzen.run()` empty.\nTo copy the default config and edit it, do the following.\n\n    from pathlib import Path\n    from event_horyzen import event_horyzen\n    \n    dest = Path("./foo/")\n    event_horyzen.copy_default_config(dest)\n\nIf you don&rsquo;t specify a destination, it will copy the file to your current working directory.\nNow, assuming you&rsquo;ve edited the config to your liking and its named `config.yml`:\n\n    from pathlib import Path\n    from event_horyzen import event_horyzen\n    \n    conf_path = Path(\'./config.yml\')\n    event_horyzen.run(conf_path)\n\n**or for multiple geodesics simulated in parallel**\n\n    from pathlib import Path\n    from event_horyzen import event_horyzen\n    \n    conf_path1 = Path(\'./config1.yml\')\n    conf_path2 = Path(\'./config2.yml\')\n    conf_path3 = Path(\'./config3.yml\')\n    \n    confs = [conf_path1, conf_path2, conf_path3]\n    \n    event_horyzen.run(confs)\n\nA unique directory under the output directory specified in the configuration file will be created in the format `<output-dir>/<date+time>_<name-of-config-file>`.\nSo, it makes sense to give your configuration files meaningful names.\nThe geodesic in both spherical and cartesian coordinates will be saved to this directory as `results.h5`.\nThe configuration file used to generate the simulation will be copied to this directory as well to ensure reproducibility.\nA basic plot of the geodesic is also created and saved in the directory as both a .PNG and a `pickle` file so that the figure can be reloaded and interacted with.\n[Example Kerr-Newman Plot](./example-kerr-newman.png)\n\nIf you want to load the pickled Matplotlib plot, you can do the following.\n\n    import pickle as pl\n    from pathlib import Path\n    \n    plot_path = Path("<path-to-plot>/basic-plot.pickle")\n    \n    with open(plot_path, "rb") as plot:\n        fig = pl.load(plot)\n    \n    # Now do whatever you want with the figure!\n    \n    fig.show()\n\nFor the 3D plotting,\n\n    from pathlib import Path\n    from event_horyzen import animated_plot as ap\n    \n    results = Path("./results.h5")\n    viz = ap.Visualizer(results)\n    viz.animation()\n\n**or for multiple geodesics on the same plot**\n\n    from pathlib import Path\n    from event_horyzen import animated_plot as ap\n    \n    results1 = Path("./results1.h5")\n    results2 = Path("./results2.h5")\n    results3 = Path("./results3.h5")\n    \n    results = [results1, results2, results3]\n    \n    viz = ap.Visualizer(results)\n    viz.animation()\n\nBy default, it puts a photon sphere for a M=1 (geometrized units) schwarzschild black hole on the plot for reference.\nThis can be turned off or modified in the call to `Visualizer()`.\n\n**Both the simulation and the plotting can be ran directly from the command line**\n\nFirst, the simulation tools.\n\n    event-horyzen -h\n\n    usage: event-horyzen [-h] [datapath ...]\n    \n    positional arguments:\n      datapath    The path(s) to the configuration file(s). Defaults to the\n                  included `config.yml` if not provided.\n    \n    options:\n      -h, --help  show this help message and exit\n\nNow, the plotting tools.\n\n    event-horyzen-plot -h\n\n    usage: event-horyzen-plot [-h] datapath [datapath ...]\n    \n    positional arguments:\n      datapath    The path(s) to the data file(s).\n    \n    options:\n      -h, --help  show this help message and exit\n\n',
    'author': 'David Wright',
    'author_email': 'davecwright@knights.ucf.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/UCF-SPS-Research-21/event-horyzen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
